/*
 * Copyright (c) 1999 Martin Jolicoeur <martinj@visuaide.com>
 *
 * Screen Driver for Linux Framebuffers, portrait mode
 */
#include <stdio.h>
#include "device.h"
#include "fb.h"
#include "genfont.h"

/* specific driver entry points*/
static int  FB_open(PSD psd);
static void FB_close(PSD psd);
static void FB_getscreeninfo(PSD psd,PSCREENINFO psi);
static void FB_setpalette(PSD psd,int first,int count,RGBENTRY *pal);
static void FB_drawpixel(PSD psd,COORD x, COORD y, PIXELVAL c);
static PIXELVAL FB_readpixel(PSD psd,COORD x, COORD y);
static void FB_drawhline(PSD psd,COORD x1, COORD x2, COORD y, PIXELVAL c);
static void FB_drawvline(PSD psd,COORD x, COORD y1, COORD y2, PIXELVAL c);
static void FB_fillrect(PSD psd,COORD x1,COORD y1,COORD x2,COORD y2,PIXELVAL c);
static void FB_blit(PSD dstpsd,COORD destx,COORD desty,COORD w,COORD h,
		PSD srcpsd,COORD srcx,COORD srcy,int op);

SCREENDEVICE	scrdev = {
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL,
	FB_open,
	FB_close,
	FB_getscreeninfo,
	FB_setpalette,
	FB_drawpixel,
	FB_readpixel,
	FB_drawhline,
	FB_drawvline,
	FB_fillrect,
	gen_getfontinfo,
	gen_gettextsize,
	gen_gettextbits,
	FB_blit,
	NULL
};

static int
FB_open(PSD psd)
{
	/* fb_init selects the proper framebuffer subdriver and
	 * inits the SCREENDEVICE struct
	 */
	if(fb_init(psd) == 0)
		return -1;	/* fail*/

	/* Flip virtual resolutions */
	psd->yvirtres = psd->xres;
	psd->xvirtres = psd->yres;

	/* Set the portrait flag */
	psd->flags |= PSF_PORTRAIT;

	return 1;		/* ok*/
}

static void
FB_close(PSD psd)
{
	fb_exit(psd);
}

static void
FB_getscreeninfo(PSD psd,PSCREENINFO psi)
{
	psi->rows = psd->yvirtres;
	psi->cols = psd->xvirtres;
	psi->planes = psd->planes;
	psi->bpp = psd->bpp;
	psi->ncolors = psd->ncolors;
	psi->pixtype = psd->pixtype;
	psi->fonts = NUMBER_FONTS;

	if(psd->yvirtres > 480) {
		/* SVGA 800x600*/
		psi->xdpcm = 33;	/* assumes screen width of 24 cm*/
		psi->ydpcm = 33;	/* assumes screen height of 18 cm*/
	} else if(psd->yvirtres > 350) {
		/* VGA 640x480*/
		psi->xdpcm = 27;	/* assumes screen width of 24 cm*/
		psi->ydpcm = 27;	/* assumes screen height of 18 cm*/
	} else {
		/* EGA 640x350*/
		psi->xdpcm = 27;	/* assumes screen width of 24 cm*/
		psi->ydpcm = 19;	/* assumes screen height of 18 cm*/
	}
}

/*
 * Set count palette entries starting from first
 */
static void
FB_setpalette(PSD psd,int first,int count,RGBENTRY *pal)
{
	fb_setpalette(psd, first, count, (void*)pal);
}

static void
FB_drawpixel(PSD psd,COORD x, COORD y, PIXELVAL c)
{
	fb_drawpixel(psd, y, psd->xvirtres-x-1, c);
}

static PIXELVAL
FB_readpixel(PSD psd,COORD x, COORD y)
{
	return fb_readpixel(psd, y, psd->xvirtres-x-1);
}

static void
FB_drawhline(PSD psd,COORD x1, COORD x2, COORD y, PIXELVAL c)
{
	/*
	 * fb drivers use type 2 line drawing, the last point is not drawn
	 */
	fb_drawvertline(psd, y, psd->xvirtres-(x2+1), psd->xvirtres-x1, c);

	/*
	 * Uncomment the following if driver doesn't support hline
	x2 = psd->xvirtres-x2-1;
	while(x2 <= (psd->xvirtres-x1-1))
		fb_drawpixel(psd, y, x2++, c);
	 */
}

static void
FB_drawvline(PSD psd,COORD x, COORD y1, COORD y2, PIXELVAL c)
{
	/*
	 * fb drivers use type 2 line drawing, the last point is not drawn
	 */
	fb_drawhorzline(psd, y1, y2+1, psd->xvirtres-x-1, c);

	/*
	 * Uncomment the following if driver doesn't support vline
	while(y1 <= y2)
		fb_drawpixel(psd, y1++, psd->xvirtres-x-1, c);
	 */
}

static void
FB_fillrect(PSD psd,COORD x1, COORD y1, COORD x2, COORD y2, PIXELVAL c)
{
	/*
	 * Call fb hline (type 2) to save size
	 */
	x2 = psd->xvirtres-x2-1;
	while(x2 <= (psd->xvirtres-x1-1))
		fb_drawhorzline(psd, y1, y2+1, x2++, c);
}

static void
FB_blit(PSD dstpsd,COORD destx,COORD desty,COORD w,COORD h,PSD srcpsd,
	COORD srcx,COORD srcy,int op)
{
	fb_blit(dstpsd, desty, dstpsd->xvirtres-destx-w, h, w,
		srcpsd, srcy, srcpsd->xvirtres-srcx-w, op);
}
